DETECTOR_NAMES = [
    "Basic Auth String",
    "Generic Password",
    "JSON Web Token",
    "Generic Terraform Variable Secret",
    "Generic Database Assignment",
    "Company Email Password",
    "Generic Password",
    "Base64 Generic High Entropy Secret",
    "Generic Database Assignment",
    "Generic CLI Option Secret",
    "Username Password",
    "Generic High Entropy Secret",
    "Base64 Basic Authentication",
    "Bearer Token",
    "Authentication Tuple",
    "Typeform API Token",
    "New Relic API Key",
    "Pingdom token v3",
    "Datadog Keys",
    "Databricks Authentication Token With Hostname",
    "GitGuardian Public Monitoring API Key",
    "GitHub Server-to-server Token",
    "Infracost API Key",
    "Facebook App Keys",
    "Firebase Cloud Messaging API Key",
    "Intercom Token",
    "Sourcegraph Access Token v1",
    "Stripe Webhook Secret",
    "GitLab Token",
    "New Relic API Service Key",
    "Eventbrite OAuth2 Token",
    "Base64 AWS SES Keys",
    "Doppler API Key",
    "Heartland API key",
    "Tailscale Pre-Authentication Key",
    "Kraken Keys",
    "Coveralls Repository Token",
    "Docker Credentials",
    "Algolia Monitoring Keys",
    "Grafana Token",
    "PackageCloud API Token",
    "Square Access Token",
    "DigitalOcean Token",
    "Sourcegraph Access Token v3",
    "Akamai API Credentials",
    "Linode Personal Access Token",
    "Scalr API Access Token",
    "FullContact Key",
    "Nylas API Key",
    "Plaid Access Token",
]
MATCH_NAMES = [
    "apikey",
    "client_id",
    "client_secret",
    "host",
    "password",
    "username",
    "token",
    "port",
    "scheme",
    "connection_uri",
    "subdomain",
    "private_key",
    "domain",
    "secret_key",
    "access_token",
    "project_id",
    "cloud_name",
    "database",
    "client_token",
    "secret_token",
    "tenant_id",
    "private_key_id",
    "integration_key",
    "azure_endpoint",
    "app_id",
    "cluster",
    "pub_key",
    "sub_key",
    "environment",
    "refresh_token",
    "organization",
    "session_token",
    "connection_string",
    "account",
    "user",
    "client_certificate",
    "client_key",
    "config_value",
]
